import jaxopt
import jax
import jax.numpy as jnp

from typing import Any
from typing import Callable
from typing import NamedTuple
from typing import Optional
from typing import Union
import os
import sys

# make higher-level lib package visible to scripts
#sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname("benchmarx"), '..')))
from jaxopt._src.proximal_gradient import ProximalGradient, ProxGradState
from jaxopt import base


class ProximalGradientCLS(jaxopt.ProximalGradient):
   linesearch_custom: jaxopt.BacktrackingLineSearch = None

   def _iter(self, iter_num, x, x_fun_val, x_fun_grad, stepsize, hyperparams_prox, args, kwargs):

    if self.linesearch_custom is not None:
      next_stepsize = self.linesearch_custom.run(init_stepsize=1.0, params=x)[0]
      next_x = self._prox_grad(x, x_fun_grad, next_stepsize, hyperparams_prox)
      return next_x, next_stepsize
    
    return super()._iter(iter_num, x, x_fun_val, x_fun_grad, stepsize, hyperparams_prox, args, kwargs)


class GradientDescentCLS(ProximalGradientCLS):

  def init_state(self, init_params: Any, *args, **kwargs) -> ProxGradState:
    return super().init_state(init_params, None, *args, **kwargs)

  def update(self, params: Any, state: NamedTuple, *args, **kwargs) -> base.OptStep:

    return super().update(params, state, None, *args, **kwargs)

  def optimality_fun(self, params, *args, **kwargs):
    return self._grad_fun(params, *args, **kwargs)

  def __post_init__(self):
    super().__post_init__()
    self.reference_signature = self.fun


from typing import Callable
import jax
import jax.numpy as jnp
import logging

from benchmarx.problem import Problem
from benchmarx.defaults import default_seed

class Rastrigin(Problem):
    """
    A class describing the Rastrigin function.

    The function is defined as:
    f(x) = A * n + sum(xi^2 - A * cos(2 * pi * xi)) for xi in x

    where A is a constant, n is the dimensionality of the problem, and x is
    the input vector.
    """
    n: int = 2      # problem dimensionality
    seed = default_seed
    def __init__(self, n: int = 2, info: str = 'Rastrigin') -> None:
        """
        Initialize the Rastrigin instance.

        Args:
            n (int, optional): Problem dimensionality.
            info (str, optional): Brief information about the problem.
        """
        self.info = info
        self.n = n
        self.x_opt = jnp.zeros(self.n)
        self.f_opt = 0.0
        super().__init__(info=info, func=self.f, x_opt=self.x_opt)
    

    def f(self, x, *args, **kwargs):
        """
        Calculate the Rastrigin function value.

        Args:
            x (Any): Input value for the function.

        Returns:
            Any: The Rastrigin function value.
        """
        if x.shape[0] != self.n:
            err_msg = f'Wrong x shape: {x.shape}. Expected: ({self.n},)'
            logging.critical(err_msg)
            exit(1)
        A = 10
        return A * self.n + jnp.sum(jnp.array([xi**2 - A * jnp.cos(2 * jnp.pi * xi) for xi in x]))



from typing import Callable


class Problem:
    """
    The base class for optimization problems.

    Attributes:
        info (str): Brief information about the problem.
        f (Callable): Target function to be optimized.
        x_opt (Any, optional): Known optimal solution.
        f_opt (Any, optional): Optimal function value corresponding to x_opt.
    """

    info: str           # Brief information about the problem, such as a common name
    f: Callable         # Target function
    x_opt: any = None   # Problem's optimizer (optional)
    f_opt: any = None   # target_func(x_opt)
    def __init__(self, info: str, func: Callable, x_opt=None) -> None:
        """
        Initialize the Problem instance.

        Args:
            info (str): Brief information about the problem.
            func (Callable): Target function to be optimized.
            x_opt (Any, optional): Known optimal solution.
        """
        self.info = info
        self.f = func
        if self.x_opt is None:
            self.x_opt = x_opt
        if self.x_opt is not None and self.f_opt is None:
            self.f_opt = self.f(self.x_opt)

    def f(self, x, *args, **kwargs):
        """
        Calculate the target function value.

        Args:
            x (Any): Input value for the target function.

        Returns:
            Any: The target function value.
        """
        return self.f(x)

    def __str__(self) -> str:
        """
        Return a string representation of the Problem.

        Returns:
            str: The problem's information.
        """
        return self.info



default_seed = 10
default_dimensionality = 10
default_log_threshold = 1e-10

default_plotly_config = {
    'displaylogo': False,
    'toImageButtonOptions': {
        'format': 'svg', # one of png, svg, jpeg, webp
        'filename': 'fmin',
        'height': None,
        'width': None,
        'scale': 1 # Multiply title/legend/axis/canvas sizes by this factor
    },
    'modeBarButtonsToRemove': ['select2d', 'lasso2d'],
    'modeBarButtonsToAdd': [
                            'drawopenpath',
                            'eraseshape'
                            ]
}


import jax.numpy as jnp


from typing import Union
from typing import Callable
from typing import Any


class State:
    iter_num: int
    stepsize: Union[float, Callable]

    def __init__(self, iter_num: int, stepsize: Union[float, Callable]) -> None:
        self.iter_num = iter_num
        self.stepsize = stepsize


class CustomOptimizer:
    """
    Base class to implement your own optimization algorithm.
    """
    params: Any         # hyperparams
    x_init: jnp.array   # initial point
    label: str          # label (for plots?)
    def __init__(self, params: Any, x_init: jnp.array, label: str) -> None:
        """
        Initialize the CustomOptimizer instance.

        Args:
            params: Hyperparameters for the optimization algorithm.
            x_init: Initial point as a NumPy array.
            label: Label for the optimization algorithm (for plots, etc.).
        """
        self.params = params
        self.x_init = x_init
        self.label = label


    def init_state(self, x_init, *args, **kwargs) -> State:
        """
        Initialize and return the initial state of the optimization algorithm.

        Args:
            x_init: Initial point for the optimization.

        Returns:
            A State object representing the initial state.
        """
        #del x_init
        return State(iter_num=1, stepsize=1.0)


    def update(self, sol, state: State) -> tuple([jnp.array, State]):
        """
        Perform an update step of the optimization algorithm and return the updated point and state.

        Args:
            sol: Current solution point.
            state: Current state of the optimization algorithm.

        Returns:
            A tuple containing the updated solution point and the updated state.
        """
        state.iter_num += 1
        return self.x_init, state


    def stop_criterion(self, sol, state: State) -> bool:
        """
        Check if the optimization should stop based on the current state.

        Args:
            sol: Current solution point.
            state: Current state of the optimization algorithm.

        Returns:
            True if the optimization should stop, False otherwise.
        """
        return True


import jax.numpy as jnp
import jax
import logging

from sklearn.datasets import load_svmlight_file, load_breast_cancer
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score

from benchmarx.model_problem import ModelProblem
from typing import Callable


class LogisticRegression(ModelProblem):
    """
    A class describing an unconstrained logistic regression problem.

    f(w) = 1/n sum_{i=1}^n f_i(w) -> min_w,
    f_i(w) = ln(1 + exp(-y_i * w.T @ x_i))

    Typical usage example:

    problem = LogisticRegression("mushrooms")
    benchmark = Benchmark(problem=problem, ...)
    result = benchmark.run()
    """

    def __init__(self, info: str=None, problem_type: str = "mushrooms", train_data_part_size: int=-1, regularizer: Callable= lambda w: 0, x_opt=None) -> None:
        """
        problem_type: The type of problem to set up. Can be one of the
                               following:
                               - 'mushrooms'
                               - 'breast_cancer'
        """
        info_str = info
        if info is None:
            info_str = f"Logistic Regression problem on {problem_type} dataset"
        super().__init__(info=info_str, x_opt=x_opt)
        self.problem_type = problem_type
        self.regularizer = regularizer

        if train_data_part_size > 0:
            self.train_data_parts = train_data_part_size

        if problem_type == "mushrooms":
            dataset = "mushrooms.txt"
            data = load_svmlight_file(dataset)
            X, y = data[0].toarray(), data[1]
            y = 2 * y - 3

            self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(
                X, y, train_size=0.8, test_size=0.2, shuffle=True
            )
            self.n_train, self.d_train = self.X_train.shape
            self.n_test, self.d_test = self.X_test.shape
        
        elif problem_type == "breast_cancer":
            cancer = load_breast_cancer()
            X = cancer.data
            y = cancer.target
            # from {0, 1} to {-1, 1}
            y = (y+1)*2-3
            self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(
                X, y, train_size=0.8, test_size=0.2, shuffle=True
            )
            self.n_train, self.d_train = self.X_train.shape
            self.n_test, self.d_test = self.X_test.shape

        else:
            raise ValueError(f"Unknown problem type: {problem_type}")

    @staticmethod
    @jax.jit
    def jitted_log_loss(w, X, y):
        """
        Calculate the logistic loss function value.

        Args:
            w (Any): Parameter vector.
            X (Any): Input features.
            y (Any): Target labels.

        Returns:
            Any: The logistic loss function value.
        """
        return jnp.mean(jnp.logaddexp(jnp.zeros(X.shape[0]), -y * (X @ w)))
    
    @staticmethod
    def log_loss(w, X, y, regularizer):
        """
        Calculate the regularized logistic loss function value.

        Args:
            w (Any): Parameter vector.
            X (Any): Input features.
            y (Any): Target labels.
            regularizer (Callable): regularizer

        Returns:
            Any: The regularized logistic loss function value.
        """
        return LogisticRegression.jitted_log_loss(w=w, X=X, y=y) + regularizer(w)
    
    def grad_log_loss_ind(self, w, ind):
        """
        The logistic regression problem has a finite sum form:
        f(w) = 1/n sum_{j=1}^n [f_j(w)],
        f_j(w) = 1/b sum_{i=1}^b l(g(w, x_i), y_i)

        where b = self.train_data_part_size,
        n*b = N -- full sample size (X.shape[0]).
        ind in range(n)
        Returns:
            1/b sum_{i=1}^b l(g(w, x_i), y_i)
        """
        X_part = self.X_train[ind * self.train_data_part_size : (ind + 1) * self.train_data_part_size]
        y_part = self.y_train[ind * self.train_data_part_size : (ind + 1) * self.train_data_part_size]
        return jax.grad(LogisticRegression.jitted_log_loss(w=w, X=X_part, y=y_part))(w)

    def accuracy(self, w, X, y):
        """
        Compute accuracy on (X, y).

        Args:
            w (Any): Parameter vector.
            X (Any): Input features.
            y (Any): Target labels.

        Returns:
            float: Accuracy score.
        """
        return accuracy_score(y, jnp.around(2 / (1 + jnp.exp(-X @ w))) - 1)

    def train_loss(self, w, *args, **kwargs):
        """
        Calculate the logistic loss function on the train part of the dataset.

        Args:
            w (Any): Parameter vector.

        Returns:
            Any: The logistic loss function value.
        """
        return LogisticRegression.log_loss(w=w, X=self.X_train, y=self.y_train, regularizer=self.regularizer)

    def test_loss(self, w, *args, **kwargs):
        """
        Calculate the logistic loss function on the test part of the dataset.

        Args:
            w (Any): Parameter vector.

        Returns:
            Any: The logistic loss function value.
        """
        return LogisticRegression.log_loss(w=w, X=self.X_test, y=self.y_test, regularizer=self.regularizer)

    def train_accuracy(self, w, *args, **kwargs):
        """
        Calculate the accuracy on the train part of the dataset.

        Args:
            w (Any): Parameter vector.

        Returns:
            float: Accuracy score.
        """
        return self.accuracy(w=w, X=self.X_train, y=self.y_train)

    def test_accuracy(self, w, *args, **kwargs):
        """
        Calculate the accuracy on the test part of the dataset.

        Args:
            w (Any): Parameter vector.

        Returns:
            float: Accuracy score.
        """
        return self.accuracy(w=w, X=self.X_test, y=self.y_test)
    
    def estimate_L(self):
        """
        Estimate the Lipschitz constant of the gradient of the logistic loss function.

        Returns:
            float: Estimated Lipschitz constant.
        """
        outer_products = jax.vmap(lambda x: jnp.outer(x, x))(self.X_train)
        sum_of_outer_products = jnp.sum(outer_products, axis=0)
        eigenvalues = jnp.linalg.eigvals(sum_of_outer_products)
        max_eigenvalue = jnp.max(eigenvalues)
        L = max_eigenvalue / self.n_train
        return float(L.real)

    def estimate_L_for_sum(self):
        """
        Estimate the Lipschitz constant of the gradient of the logistic loss function
        for each part of datasets and returns maximum.

        Returns:
            float: Estimated Lipschitz constant for sum.
        """
        Ls = []
        n = self.n_train // self.tr


from typing import Callable

from abc import abstractmethod
from benchmarx.problem import Problem


class ModelProblem(Problem):
    """
    Abstract base class describing a learning problem.

    Attributes:
        info (str): Information about the problem.
        x_opt (Optional): Optimal solution (if available).
    """

    def __init__(self, info: str, x_opt=None) -> None:
        super().__init__(info, self.train_loss, x_opt)

    @abstractmethod
    def train_loss(self, w, *args, **kwargs):
        """
        Abstract method for computing the training loss.
        """
        pass

    @abstractmethod
    def test_loss(self, w, *args, **kwargs):
        """
        Abstract method for computing the test loss.
        """
        pass

    @abstractmethod
    def train_accuracy(self, w, *args, **kwargs):
        """
        Abstract method for computing the training accuracy.
        """
        pass

    @abstractmethod
    def test_accuracy(self, w, *args, **kwargs):
        """
        Abstract method for computing the test accuracy.
        """
        pass




import logging
from typing import List

available_built_in_methods = [
    'GRADIENT_DESCENT',
    'BFGS',
    'LBFGS',
    'ArmijoSGD',
    'PolyakSGD',
    'NonlinearCG'
]

def check_method(methods: List[str]):
    """
    Check if the provided methods are supported built-in methods.

    Args:
        methods (List[str]): List of method names.

    Returns:
        bool: True if all methods are supported built-in methods, False otherwise.
    """
    for method in methods:
        for avail_method in available_built_in_methods:
            if method.startswith(avail_method):
                return True
    logging.critical(f'Unsupported built in method \'{method}\'. Available built in methods: {available_built_in_methods}')
    return False




import jaxopt
import jax
import jax.numpy as jnp

import time
import logging
from typing import List, Dict

from benchmarx.problem import Problem
from benchmarx.model_problem import ModelProblem
import benchmarx.methods as _methods
import benchmarx.metrics as _metrics
from benchmarx.benchmark_result import BenchmarkResult
from benchmarx.custom_optimizer import CustomOptimizer

from benchmarx.ProxGD_custom_linesearch import GradientDescentCLS
from benchmarx.defaults import default_seed


class Benchmark:
    """
    A class that provides benchmarking of different optimization methods on a given problem.

    Note: nfev, njev, nhev metrics aoutomaticly disable jax.jit.
    """

    runs: int = 1  # the number of runs of each method
    problem: Problem = None  # Problem to solve
    methods: List[Dict[str, Dict[str, any]]] = None  # Methods for benchmarking
    available_built_in_methods: List[str] = None  # method's keywords.
    # If you want to call a method from the jaxopt,
    # the name of the method must begin with one of these keywords.
    str_metrics_to_track: List[str]  # subset of _metrics.Metrics.metrics_to_track
    custom_metrics_to_track: List[_metrics.CustomMetric]  # your own custom metrics
    aval_linesearch_str = ["armijo", "goldstein", "strong-wolfe", "wolfe"]

    def __init__(
        self,
        problem: Problem,
        methods: List[Dict[str, Dict[str, any]]],
        metrics: List[str | _metrics.CustomMetric], # metrics to track
        runs: int = 1,
    ) -> None:
        """
        Initialize the Benchmark instance.

        Args:
            problem: A Problem class object (or inheritor).
            methods: A list of dictionaries with method names and their corresponding parameters.
            metrics: A list of metrics to track.
            runs: Number of runs for each method.
        """
        self.runs = runs
        self.problem = problem
        self.methods = methods
        self.available_built_in_methods = _methods.available_built_in_methods
        
        self.str_metrics_to_track = [metric for metric in metrics if isinstance(metric, str)]
        _metrics.Metrics.check_metrics_to_track(
            metrics_to_check=self.str_metrics_to_track
        )

        self.str_metrics_to_track = _metrics.Metrics.fix_metrics_to_track(
            metrics_to_fix=self.str_metrics_to_track
        )
        self.custom_metrics_to_track = [
            metric for metric in metrics if isinstance(metric, _metrics.CustomMetric)
        ]

        self.nfev_global = 0    # number of objective function evaluations on the current iteration
        self.njev_global = 0    # number of objective function gradient evaluations on the current iteration
        self.nhev_global = 0    # number of objective function hessian evaluations on the current iteration
        


    def _check_linesearch(self, ls_str: str, method: str):
        """
        Check if a given line search method is valid for a specific optimization method.

        Args:
            ls_str: Line search method name.
            method: Optimization method name.

        Returns:
            True if the line search is valid for the optimization method, False otherwise.
        """
        # TODO: 'steepest' for QuadraticProblem!
        if method.startswith("GRADIENT_DESCENT"):
            return ls_str in ["armijo", "goldstein", "strong-wolfe", "wolfe"]
        elif method.startswith("BFGS") or method.startswith("LBFGS"):
            return ls_str in [
                "armijo",
                "goldstein",
                "strong-wolfe",
                "wolfe",
            ] or ls_str in ["backtracking", "zoom", "hager-zhang"]

        return False

    def __run_solver(
        self, 
        solver, 
        x_init,
        *args, 
        **kwargs
    ) -> Dict[str, List[any]]:
        """
        Run an optimization solver and collect metrics.

        Args:
            solver: The solver object to be run.
            x_init: Initial solution.
            *args: Additional arguments for the solver.
            **kwargs: Additional keyword arguments for the solver.

        Returns:
            A dictionary containing collected metrics.
        """
        # set nfev, njev and nhev to 0 at the begining of method
        self.nfev_global = 0
        self.njev_global = 0
        self.nhev_global = 0

        count_calls = "nfev" in self.str_metrics_to_track or "njev" in self.str_metrics_to_track or "nhev" in self.str_metrics_to_track

        custom_method_flag = issubclass(type(solver), CustomOptimizer)
        # cls = hasattr(solver, 'linesearch_custom')
        result = dict()
        state = solver.init_state(x_init, *args, **kwargs)
        sol = x_init
        if custom_method_flag and sol is None:
            sol = solver.x_init

        x_prev = sol

        @jax.jit
        def jitted_update(sol, state):
            # logging.warning(f'Function \"jitted_update\" doesn\'t use jax.jit.')
            return solver.update(sol, state, *args, **kwargs)

        def update(sol, state):
            return solver.update(sol, state, *args, **kwargs)

        def stop_criterion(err, tol):
            return err < tol

        tol = 0

        if not custom_method_flag and "tol" in kwargs:
            tol = kwargs["tol"]

        start_time = time.time()


        iters_total = 0
        for i in range(solver.maxiter + 1):
            if i > 0:
                if not custom_method_flag and stop_criterion(state.error, tol):
                    break
                if custom_method_flag and solver.stop_criterion(sol, state):
                    break
                iters_total = i + 1

            if isinstance(sol, float):
                sol = jnp.array([sol])
            
            # metrics to track as method works: ["x", "nfev", "nhev", "njev", "time"]
            # "x" is always in self.str_metrics_to_track
            if not "x" in result:
                result["x"] = [sol]
            else:
                result["x"].append(sol)

            # "nit" is always in self.str_metrics_to_track
            if not "nit" in result:
                result["nit"] = [1]
            else:
                result["nit"][0] += 1
            

            if custom_method_flag or count_calls:
                sol, state = update(sol, state)
            else:
                sol, state = jitted_update(sol, state)
            
            if "nfev" in self.str_metrics_to_track:
                if "nfev" not in result:
                    result["nfev"] = [self.nfev_global]
                else:
                    result["nfev"].append(self.nfev_global)

            if "njev" in self.str_metrics_to_track:
                if "njev" not in result:
                    result["njev"] = [self.njev_global]
                else:
                    result["njev"].append(self.njev_global)

            if "nhev" in self.str_metrics_to_track:
                # in progress
                if "nhev" not in result:
                    result["nhev"] = [self.nhev_global]
                else:
                    result["nhev"].append(self.nhev_global)
            
            if "time" in self.str_metrics_to_track:
                if not "time" in result:
                    result["time"] = [time.time() - start_time]
                else:
                    result["time"].append(time.time() - start_time)
            
            x_prev = sol

        # metrics to track at the end
        for i in range(iters_total):
            sol = result["x"][i]
            # objective function values
            if "f" in self.str_metrics_to_track:
                if not "f" in result:
                    result["f"] = [self.problem.f(sol)]
                else:
                    result["f"].append(self.problem.f(sol))
            
            # gradient of the objective function
            if "grad" in self.str_metrics_to_track:
                grad_val = self.problem.grad(sol) if hasattr(self.problem, "grad") else jax.grad(self.problem.f)(sol)
                if not "grad" in result:
                    result["grad"] = [grad_val]
                else:
                    result["grad"].append(grad_val)
            
            # for ModelProblem problem it's necessary to track:
            # train_loss (aka problem.f)
            # test_loss
            # train_accuracy
            # test_accuracy
            if isinstance(self.problem, ModelProblem):
                # train_loss metric tracking
                train_loss_val = None
                if "f" in self.str_metrics_to_track:
                    train_loss_val = result["f"][-1]
                else:
                    train_loss_val = self.problem.train_loss(sol)

                if not "train_loss" in result:
                    result["train_loss"] = [train_loss_val]
                else:
                    result["train_loss"].append(train_loss_val)

                # test_loss metric tracking
                test_loss_val = self.problem.test_loss(sol)
                if not "test_loss" in result:
                    result["test_loss"] = [test_loss_val]
                else:
                    result["test_loss"].append(test_loss_val)

                # train_accuracy metric tracking
                train_accuracy_val = self.problem.train_accuracy(sol)
                if not "train_accuracy" in result:
                    result["train_accuracy"] = [train_accuracy_val]
                else:
                    result["train_accuracy"].append(train_accuracy_val)
                
                # test_accuracy metric tracking
                test_accuracy_val = self.problem.test_accuracy(sol)
                if not "test_accuracy" in result:
                    result["test_accuracy"] = [test_accuracy_val]
                else:
                    result["test_accuracy"].append(test_accuracy_val)
                    

            # custom metrics moment
            for custom_metric in self.custom_metrics_to_track:
                if i % custom_metric.step == 0:
                    if not custom_metric.label in result:
                        result[custom_metric.label] = [custom_metric.func(sol)]
                    else:
                        result[custom_metric.label].append(custom_metric.func(sol))

        return result
    
    def traced_objective_function(self, x):
        self.nfev_global += 1
        return self.problem.f(x)
    
    def traced_gradient_function(self, x):
        self.njev_global += 1
        if hasattr(self.problem, "grad"):
            return self.problem.grad(x)
        return jax.grad(self.problem.f)(x)
    
    def tracked_objective_and_gradient(self, x, *args, **kwargs):
        return self.traced_objective_function(x), self.traced_gradient_function(x)

    def run(self) -> BenchmarkResult:
        """
        Run benchmarking for the specified optimization methods on the given problem.

        Returns:
            A BenchmarkResult object containing the benchmarking results.
        """
        res = BenchmarkResult(
            problem=self.problem, 
            methods=list()
        )
        data = dict()
        data[self.problem] = dict()
        # methods: list[dict[method(str) : dict[str:any]]]
        for item in self.methods:
            for method, params in item.items():
                if not isinstance(params, CustomOptimizer):
                    # data: dict[Problem, dict[method(str), dict[str, list[any]]]]

                    # ======= custom line search =======
                    # A class jaxopt.BacktrackingLineSearch object or str is expected.
                    # For Gradient Descent: params['linesearch'] in ['wolfe', 'strong-wolfe', 'armijo', 'goldstein']
                    # For (L)BFGS: params['linesearch'] must be str from
                    # ['backtracking', 'zoom', 'hager-zhang'] or ['wolfe', 'strong-wolfe', 'armijo', 'goldstein'].
                    if not _methods.check_method([method]):
                        continue

                    cls = "linesearch" in params

                    if method.startswith("GRADIENT_DESCENT"):
                        logging.info("Default gradient descent")
                        res.methods.append(method)
                        x_init = None
                        label = "jaxopt.GradientDescent"
                        if hasattr(self.problem, "sedd"):
                            seed = str(self.problem.seed)
                        else:
                            seed = str(default_seed)
                        if "x_init" in params:
                            x_init = params["x_init"]
                            params.pop("x_init")
                        if "label" in params:
                            label = params["label"]
                            params.pop("label")
                        if "seed" in params:
                            seed = params["seed"]
                            params.pop("seed")
                        runs_dict = dict()
                        solver = None
                        if cls:
                            ls = params["linesearch"]
                            params.pop("linesearch")
                            if "condition" in params:
                                condition = params["condition"]
                                params.pop("condition")
                            if isinstance(ls, str):
                                if ls == "backtracking":
                                    if condition in [
                                        "wolfe",
                                        "strong-wolfe",
                                        "armijo",
                                        "goldstein",
                                    ]:
                                        ls_obj = jaxopt.BacktrackingLineSearch(
                                            fun=self.tracked_objective_and_gradient,
                                            value_and_grad=True,
                                            maxiter=20,
                                            condition=condition,
                                            decrease_factor=0.8,
                                        )
                                    else:
                                        err_msg = f"Unknown condition {condition}"
                                        logging.critical(err_msg)
                                        exit(1)
                                elif ls == "hager-zhang":
                                    ls_obj = jaxopt.HagerZhangLineSearch(
                                        fun=self.tracked_objective_and_gradient,
                                        value_and_grad=True,
                                    )
                                else:
                                    err_msg = f"Unknown line search {ls}"
                                    logging.critical(err_msg)
                                    exit(1)
                                solver = GradientDescentCLS(
                                    fun=self.tracked_objective_and_gradient,
                                    value_and_grad=True, 
                                    **params
                                )
                                solver.linesearch_custom = ls_obj
                            elif isinstance(ls, jaxopt.BacktrackingLineSearch):
                                solver = GradientDescentCLS(
                                    fun=self.tracked_objective_and_gradient,
                                    value_and_grad=True,
                                    **params
                                )
                                solver.linesearch_custom = ls
                            else:
                                err_msg = f"Unknown linesearch {ls}"
                                logging.critical(err_msg)
                                exit(1)
                        else:
                            solver = jaxopt.GradientDescent(
                                fun=self.tracked_objective_and_gradient,
                                value_and_grad=True,
                                **params
                            )

                        for run in range(self.runs):
                            if run % 10 == 0:
                                logging.info(f"#{run} run...")
                            sub = self.__run_solver(
                                solver=solver,
                                x_init=x_init,
                                **params,
                            )
                            runs_dict[f"run_{run}"] = sub
                        params["x_init"] = x_init
                        params["label"] = label
                        params["seed"] = seed
                        data[self.problem][method] = {
                            "hyperparams": params,
                            "runs": runs_dict,
                        }

                    elif method.startswith("BFGS"):
                        logging.info("BFGS (jaxopt built-in)")
                        res.methods.append(method)
                        x_init = None
                        label = "jaxopt.BFGS"
                        if hasattr(self.problem, "seed"):
                            seed = str(self.problem.seed)
                        else:
                            seed = str(default_seed)
                        if "x_init" in params:
                            x_init = params["x_init"]
                            params.pop("x_init")
                        if "label" in params:
                            label = params["label"]
                            params.pop("label")
                        if "seed" in params:
                            seed = params["seed"]
                            params.pop("seed")
                        runs_dict = dict()
                        soler = None
                        if cls:
                            new_linesearch = "zoom"
                            new_condition = "stron-wolfe"
                            ls = params["linesearch"]
                            params.pop("linesearch")
                            cond = ""
                            if "condition" in params:
                                cond = params["condition"]
                                params.pop("condition")
                            if isinstance(ls, str) and self._check_linesearch(
                                ls, method
                            ):
                                if ls in ["backtracking", "zoom", "hager-zhang"]:
                                    new_linesearch = ls
                                else:
                                    err_msg = f"Unknown line search '{ls}'. zoom line search will be used instead of {ls}."
                                    logging.warning(err_msg)
                                if cond in [
                                    "wolfe",
                                    "strong-wolfe",
                                    "armijo",
                                    "goldstein",
                                ]:
                                    new_condition = cond
                                else:
                                    err_msg = f"Unknown condition '{cond}'. strong-wolfe condition will be used instead if {cond}"
                                    logging.warning(err_msg)
                            else:
                                err_msg = f"For BFGS parameter 'linesearch' must be string from {['wolfe', 'strong-wolfe', 'armijo', 'goldstein']}(condition) or {['backtracking', 'zoom', 'hager-zhang']} (linesearch)"
                                logging.critical(err_msg)

                            solver = jaxopt.BFGS(
                                fun=self.tracked_objective_and_gradient,
                                value_and_grad=True,
                                linesearch=new_linesearch,
                                condition=new_condition,
                                **params,
                            )
                        else:
                            solver = jaxopt.BFGS(
                                fun=self.tracked_objective_and_gradient,
                                value_and_grad=True,
                                **params)

                        for run in range(self.runs):
                            if run % 10 == 0:
                                logging.info(f"#{run} run...")
                            sub = self.__run_solver(
                                solver=solver,
                                x_init=x_init,
                                **params,
                            )
                            runs_dict[f"run_{run}"] = sub
                        params["x_init"] = x_init
                        params["label"] = label
                        params["seed"] = seed
                        data[self.problem][method] = {
                            "hyperparams": params,
                            "runs": runs_dict,
                        }

                    elif method.startswith("LBFGS"):
                        logging.info("LBFGS (jaxopt built-in)")
                        res.methods.append(method)
                        x_init = None
                        label = "jaxopt.LBFGS"
                        if hasattr(self.problem, "seed"):
                            seed = str(self.problem.seed)
                        else:
                            seed = str(default_seed)
                        if "x_init" in params:
                            x_init = params["x_init"]
                            params.pop("x_init")
                        if "label" in params:
                            label = params["label"]
                            params.pop("label")
                        if "seed" in params:
                            seed = params["seed"]
                            params.pop("seed")
                        runs_dict = dict()
                        for run in range(self.runs):
                            if cls:
                                new_linesearch = "zoom"
                                new_condition = "stron-wolfe"
                                ls = params["linesearch"]
                                params.pop("linesearch")
                                cond = ""
                                if "condition" in params:
                                    cond = params["condition"]
                                    params.pop("condition")
                                if isinstance(ls, str) and self._check_linesearch(
                                    ls, method
                                ):
                                    if ls in ["backtracking", "zoom", "hager-zhang"]:
                                        new_linesearch = ls
                                    if cond in [
                                        "wolfe",
                                        "strong-wolfe",
                                        "armijo",
                                        "goldstein",
                                    ]:
                                        new_condition = cond
                                    else:
                                        err_msg = f"Unknown line search '{ls}', {cond}"
                                        logging.critical(err_msg)
                                        exit(1)
                                else:
                                    err_msg = f"For LBFGS parameter 'linesearch' must be string from {['wolfe', 'strong-wolfe', 'armijo', 'goldstein']}(condition) or {['backtracking', 'zoom', 'hager-zhang']} (linesearch)"
                                    logging.critical(err_msg)

                                solver = jaxopt.LBFGS(
                                    fun=self.tracked_objective_and_gradient,
                                    value_and_grad=True,
                                    linesearch=new_linesearch,
                                    condition=new_condition,
                                    **params,
                                )
                            else:
                                solver = jaxopt.LBFGS(
                                    fun=self.tracked_objective_and_gradient,
                                    value_and_grad=True, 
                                    **params)
                            sub = self.__run_solver(
                                solver=solver,
                                x_init=x_init,
                                **params,
                            )
                            runs_dict[f"run_{run}"] = sub
                        params["x_init"] = x_init
                        params["label"] = label
                        params["seed"] = seed
                        data[self.problem][method] = {
                            "hyperparams": params,
                            "runs": runs_dict,
                        }

                    elif method.startswith("ArmijoSGD"):
                        logging.info("ArmijoSGD (jaxopt built-in)")
                        res.methods.append(method)
                        x_init = None
                        label = "jaxopt.ArmijoSGD"
                        if hasattr(self.problem, "seed"):
                            seed = str(self.problem.seed)
                        else:
                            seed = str(default_seed)
                        if "x_init" in params:
                            x_init = params["x_init"]
                            params.pop("x_init")
                        if "label" in params:
                            label = params["label"]
                            params.pop("label")
                        if "seed" in params:
                            seed = params["seed"]
                            params.pop("seed")
                        runs_dict = dict()
                        solver = jaxopt.ArmijoSGD(
                            fun=self.tracked_objective_and_gradient,
                            value_and_grad=True, 
                            **params)
                        for run in range(self.runs):
                            if run % 10 == 0:
                                logging.info(f"#{run} run...")
                            sub = self.__run_solver(
                                solver=solver,
                                x_init=x_init,
                                **params,
                            )
                            runs_dict[f"run_{run}"] = sub
                        params["x_init"] = x_init
                        params["label"] = label
                        params["seed"] = seed
                        data[self.problem][method] = {
                            "hyperparams": params,
                            "runs": runs_dict,
                        }

                    elif method.startswith("PolyakSGD"):
                        logging.info("PolyakSGD (jaxopt built-in)")
                        res.methods.append(method)
                        x_init = None
                        label = "jaxopt.PolyakSGD"
                        if hasattr(self.problem, "seed"):
                            seed = str(self.problem.seed)
                        else:
                            seed = str(default_seed)
                        if "x_init" in params:
                            x_init = params["x_init"]
                            params.pop("x_init")
                        if "label" in params:
                            label = params["label"]
                            params.pop("label")
                        if "seed" in params:
                            seed = params["seed"]
                            params.pop("seed")
                        runs_dict = dict()
                        solver = jaxopt.PolyakSGD(
                            fun=self.tracked_objective_and_gradient,
                            value_and_grad=True, 
                            **params)
                        for run in range(self.runs):
                            if run % 10 == 0:
                                logging.info(f"#{run} run...")
                            sub = self.__run_solver(
                                solver=solver,
                                x_init=x_init,
                                **params,
                            )
                            runs_dict[f"run_{run}"] = sub
                        params["x_init"] = x_init
                        params["label"] = label
                        params["seed"] = seed
                        data[self.problem][method] = {
                            "hyperparams": params,
                            "runs": runs_dict,
                        }
                    elif method.startswith("NonlinearCG"):
                        logging.info("NonlinearCG (jaxopt built-in)")
                        res.methods.append(method)
                        x_init = None
                        label = "jaxopt.NonlinearCG"
                        if hasattr(self.problem, "seed"):
                            seed = str(self.problem.seed)
                        else:
                            seed = str(default_seed)
                        if "x_init" in params:
                            x_init = params["x_init"]
                            params.pop("x_init")
                        if "label" in params:
                            label = params["label"]
                            params.pop("label")
                        if "seed" in params:
                            seed = params["seed"]
                            params.pop("seed")
                        runs_dict = dict()
                        solver = jaxopt.NonlinearCG(
                            fun=self.tracked_objective_and_gradient,
                            value_and_grad=True, 
                            **params)
                        for run in range(self.runs):
                            if run % 10 == 0:
                                logging.info(f"#{run} run...")
                            sub = self.__run_solver(
                                solver=solver,
                                x_init=x_init,
                                **params,
                            )
                            runs_dict[f"run_{run}"] = sub
                        params["x_init"] = x_init
                        params["label"] = label
                        params["seed"] = seed
                        data[self.problem][method] = {
                            "hyperparams": params,
                            "runs": runs_dict,
                        }

                else:
                    # params is custom_solver object now
                    custom_solver = params
                    logging.info("Custom method")
                    res.methods.append(method)
                    runs_dict = dict()
                    x_init = custom_solver.x_init
                    for run in range(self.runs):
                        if run % 10 == 0:
                            logging.info(f"#{run} run...")
                        sub = self.__run_solver(
                            solver=custom_solver,
                            x_init=x_init
                        )
                        runs_dict[f"run_{run}"] = sub
                    params_to_write = custom_solver.params
                    if "x_init" not in params_to_write:
                        params_to_write["x_init"] = x_init
                    params_to_write["label"] = custom_solver.label
                    if hasattr(self.problem, "seed"):
                        params_to_write["seed"] = self.problem.seed
                    else:
                        params_to_write["seed"] = default_seed
                    data[self.problem][method] = {
                        "hyperparams": params_to_write,
                        "runs": runs_dict,
                    }

        res.data = data
        return res



import logging
from typing import List, Callable


class Metrics:
    """
    Class to define and manage various metrics for optimization methods.
    [1]	x	    every json-file must include this metric
    [2]	f	    objective function value
    [3]	grad	gradient of the objective function
    [4]	nit 	total number of iteration -- every json-file must include
                        this metric!
    [5]	nfev 	number of the objective function evaluations
    [6]	njev 	number of the objective function's gradient evaluations
    [7]	nhev	number of the objective function's hessian evaluation
    [8] time    At the k-th iteration, the value of the time metric is the
                time elapsed from the start of the method to the k-th iteration.

    Metrics to plot:
    (to plot a metric, it is not necessary for the corresponding
    metric to be tracked over the course of the method.
    Each of these metrics can be calculated along the method trajectory.)

        label		        value
    [1]	x_gap ............. ||x-x_opt||
    [2] 	f ............. f
    [3]	f_gap ............. |f-f_opt|
    [4]	grad_norm ......... ||grad||
    [5]	x_norm ............ ||x||
    [6] relative_x_gap .... ||x-x_opt||/||x_opt||, if x_opt != 0
    [7] relative_f_gap .... ||f-f_opt||/||f_opt||, if f_opt != 0

    Note: if one of the compulsory metrics to track is not
    contained in the list of metrics to track (it is set when
    the benchmark object is created) then this compulsory metric
    will be added to metrics to track.
    """

    compulsory_metrics_to_track = ["x", "nit"]
    metrics_to_track = ["x", "f", "grad", "nit", "nfev", "njev", "nhev", "time"]
    metrics_to_plot = ["x_gap", "f", "f_gap", "grad_norm", "x_norm", "relative_x_gap", "relative_f_gap"]
    metrics_to_track_now = ["x", "nfev", "nhev", "njev", "time"]
    model_metrics_to_plot = ["train_loss", "test_loss", "train_accuracy", "test_accuracy"]

    def __init__(self) -> None:
        pass

    @staticmethod
    def check_metrics_to_track(metrics_to_check: List[str]):
        """
        Check if provided metrics to track are in the default metrics_to_track list.
        """
        for metric in metrics_to_check:
            if metric not in Metrics.metrics_to_track:
                logging.warning(
                    msg=f"Metric '{metric}' is not contained in default metrics_to_track list. Use CustomMetric instead to specify your own metric."
                )

    @staticmethod
    def fix_metrics_to_track(metrics_to_fix: List[str]) -> List[str]:
        """
        Add compulsory metrics and remove metrics that are
        not contained in self.metrics_to_track.
        """
        fixed_metrics = Metrics.compulsory_metrics_to_track
        for metric in metrics_to_fix:
            if metric in Metrics.metrics_to_track:
                fixed_metrics.append(metric)

        return fixed_metrics

    @staticmethod
    def check_metrics_to_plot(metrics_to_check: List[str]):
        """
        Check if provided metrics to plot are in the default metrics_to_plot list.
        """
        for metric in metrics_to_check:
            if metric not in Metrics.metrics_to_plot:
                logging.warning(
                    msg=f"Metric '{metric}' is not contained in default metrics_to_plot list. Use CustomMetric instead to specify your own metric."
                )


class CustomMetric:
    """
    Class to define custom metrics for optimization methods.    
    
    It is assumed that the metric is calculated using points from
    the method trajectory, i.e. function func is real-valued and
    takes as an argument a point from R^d, where d is dimensionality
    of the problem.
    The step parameter is responsible for the frequency of func
    calculation.
    self.func will be calculated at each iteration whose number
    is a multiple of self.step.
    
    Attributes:
        func (Callable): The function to compute the custom metric.
        label (str): The label for the custom metric.
        step (int): The step parameter for frequency of computation.
    """

    def __init__(self, func: Callable, label: str, step: int = 1) -> None:
        self.func = func
        self.label = label
        self.step = step

    def __str__(self) -> str:
        return self.label



from benchmarx.problem import Problem
from benchmarx.defaults import default_seed
import jax.numpy as jnp
from jax import random
import logging
from sklearn.datasets import load_wine
import pandas as pd

class LinearLeastSquares(Problem):
    """
    A class describing an unconstrained linear least squares problem.

    f(x) = 1/2*||Ax - b||_2^2 -> min_x

    Typical usage example:

    problem = LinearLeastSquares("random", m=3, n=3)
    benchmark = Benchmark(problem=problem, ...)
    result = benchmark.run()
    """

    def __init__(self, problem_type: str, 
        A: jnp.array=None, b: jnp.array=None, x_opt: jnp.array=None, 
        m: int=None, n: int=None, seed: int=default_seed,
        reduce: bool=False) -> None:
        """
        Initializes the LinearLeastSquares problem.

        Args:
            problem_type: The type of problem to set up. Can be one of the
                               following:
                               - 'random'
                               - 'boston'
                               - 'wine'
                               - 'custom'
            A (optional): Matrix A for the custom problem. Defaults to None.
            b (optional): Vector b for the custom problem. Defaults to None.
            x_opt(optional): Optimal solution for the custom problem. 
                                Defaults to None.
            m (optional): Dimension of the data for random types. 
                               Defaults to None.
            n (optional): Dimension of the problem for random types. 
                               Defaults to None.
            seed (optional): Seed for RNG. Defaults to benchmarx 
                                default_seed.
        Raises:
            ValueError: If the provided problem_type is not recognized.
            ValueError: If required parameters (A, b for 'custom', 
                m, n for 'random') are not provided.
        """
        super().__init__(
            info=f"Linear Least Squares problem on {problem_type} dataset", 
            func=self.f
            )

        self.seed = seed
        self.reduce = reduce

        if problem_type == 'custom':
            if A is None or b is None:
                raise ValueError("For custom problem type, "
                    "A and b must be provided.")
            self.A = A
            self.b = b
            if x_opt is None:
                logging.debug("Solving custom linear system with "
                                "jnp.linalg.lstsq ")
                self.x_opt = jnp.linalg.lstsq(A, b)[0]
            else:
                self.x_opt = x_opt

        elif problem_type == 'random':
            if m is None or n is None:
                raise ValueError("For random problem types, "
                    "m and n must be provided.")
            self.A, self.b, self.x_opt = self._generate_data(m, n)
        elif problem_type in ['boston', 'wine']:
            self.A, self.b, self.x_opt = self._load_data(problem_type)
        else:
            raise ValueError(f"Unknown problem type: {problem_type}")
        
        self.f_opt = self.f(self.x_opt)
        
    def _generate_data(self, m, n):
        """
        Generates random data for the linear system.

        Args:
            m (int): Dimension of the data.
            n (int): Dimension of the problem.

        Returns:
            A: Matrix A for the problem.
            b: Vector b for the problem.
            x_opt: Optimal solution for the problem
        """

        # Log system type
        if m < n:
            logging.debug("Generating underdetermined system.")
        elif m == n:
            logging.debug("Generating square system.")
        elif m > n:
            logging.debug("Generating overdetermined system.")

        RNG = random.PRNGKey(self.seed)

        A = random.normal(RNG, (m, n))
        x_opt = random.normal(RNG, (n,))
        b = jnp.dot(A, x_opt)

        return A, b, x_opt

    def _load_data(self, problem_type):
        """
        Loads problem data from available datasets.

        Args:
            problem_type (str): Name of the dataset to load.

        Returns:
            A (jnp.array): Matrix A for the problem.
            b (jnp.array): Vector b for the problem.
            x_opt (jnp.array): Optimal solution vector.
        """
        if problem_type == 'boston':
            data_url = "http://lib.stat.cmu.edu/datasets/boston"
            raw_df = pd.read_csv(data_url, sep="\s+", skiprows=22,header=None)
            data = jnp.hstack([raw_df.values[::2, :], raw_df.values[1::2, :2]])
            target = raw_df.values[1::2, 2]
            logging.debug("Loading boston house-prices datase")
            A = data
            b = target
        elif problem_type == 'wine':
            logging.debug("Loading Wine dataset")
            data = load_wine()
            A = data.data
            b = data.target
        else:
            raise ValueError(f"Unknown dataset name: {problem_type}")

        # Compute optimal solution using least squares
        x_opt = jnp.linalg.lstsq(A, b)[0]

        return A, b, x_opt

    def f(self, x: jnp.array, *args, **kwargs):
        """
        Function to minimize.
        """
        x = jnp.array(x)
        if self.reduce:
            m, n = self.A.shape 
            return 1/(2*m)*jnp.linalg.norm(self.A@x - self.b)**2
        else:
            return 0.5*jnp.linalg.norm(self.A@x - self.b)**2


### Imports
import jax
import jax.numpy as jnp               # JAX NumPy

from flax import linen as nn          # The Linen API
from flax.training import train_state
import optax                          # The Optax gradient processing and 
                                      # optimization library

import numpy as np                    # Ordinary NumPy
import tensorflow_datasets as tfds    # TFDS for MNIST
from tqdm.auto import tqdm
from matplotlib import pyplot as plt

import time
import json
import logging
import sys
import os
from typing import List, Dict

from benchmarx.plotter import Plotter

#sys.path.insert(1, os.path.abspath(os.path.join(os.path.dirname("benchmarx"), "..")))

### Model
class CNN(nn.Module):
    @nn.compact
    # Provide a constructor to register a new parameter 
    # and return its initial value
    def __call__(self, x):
        x = nn.Conv(features=32, kernel_size=(3, 3))(x)
        x = nn.relu(x)
        x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
        x = nn.Conv(features=64, kernel_size=(3, 3))(x)
        x = nn.relu(x)
        x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
        x = x.reshape((x.shape[0], -1)) # Flatten
        x = nn.Dense(features=256)(x)
        x = nn.relu(x)
        x = nn.Dense(features=10)(x)    # There are 10 classes in FashionMNIST
        return x


model = CNN()

class NeuralNetwokTraining:
    train_ds = None
    test_ds = None
    config: Dict = None
    aval_methods = ['sgd', 'adam', 'novograd', 'adagrad']
    method: str = None
    def __init__(self, config = None, method: str = 'sgd') -> None:
        global model
        model = CNN()
        if method not in self.aval_methods:
            logging.critical(f'Wrong method {method}. Available methods: {self.aval_methods}.')
            exit(1)
        self.method = method
        if config is not None:
            self.config = config
        else:
            self.config = {
                        "N_EPOCHS": 20,
                        "BATCH_SIZE": 1128,
                        "Dataset": "FashionMNIST",
                        "seed": 282,
                        "LEARNING_RATE": 1e-1
                        }
        self.train_ds, self.test_ds = self._get_datasets()

    def _get_datasets(self):
        ds_builder = tfds.builder('fashion_mnist')
        ds_builder.download_and_prepare()
        # Split into training/test sets
        train_ds = tfds.as_numpy(ds_builder.as_dataset(split='train', 
                                                    batch_size=-1))
        test_ds = tfds.as_numpy(ds_builder.as_dataset(split='test', 
                                                    batch_size=-1))
        # Convert to floating-points
        train_ds['image'] = jnp.float32(train_ds['image']) / 255.0
        test_ds['image'] = jnp.float32(test_ds['image']) / 255.0
        return train_ds, test_ds
    
    ### Utilities
    def compute_metrics(self, logits, labels):
        loss = jnp.mean(optax.softmax_cross_entropy(logits, 
                                            jax.nn.one_hot(labels, num_classes=10)))
        accuracy = jnp.mean(jnp.argmax(logits, -1) == labels)
        metrics = {
            "loss": loss,
            "accuracy": accuracy
        }
        return metrics

    #@jax.jit
    def train_step(self, state, batch):
        def loss(params):
            logits = model.apply({'params': params}, batch['image'])
            loss = jnp.mean(optax.softmax_cross_entropy(
                logits=logits, 
                labels=jax.nn.one_hot(batch['label'], num_classes=10)))
            return loss, logits
        grad_fn = jax.value_and_grad(loss, has_aux=True)
        (_, logits), grads = grad_fn(state.params)
        state = state.apply_gradients(grads=grads)
        metrics = self.compute_metrics(logits, batch['label'])
        return state, metrics

    def train_epoch(self, state, train_ds, batch_size, epoch, rng):
        train_ds_size = len(train_ds['image'])
        steps_per_epoch = train_ds_size // batch_size

        perms = jax.random.permutation(rng, len(train_ds['image']))
        perms = perms[:steps_per_epoch * batch_size]  # Skip an incomplete batch
        perms = perms.reshape((steps_per_epoch, batch_size))

        batch_metrics = []
        total_time = 0
        iters = 0
        for perm in perms:
            batch = {k: v[perm, ...] for k, v in train_ds.items()}
            start = time.time()
            state, metrics = self.train_step(state, batch)
            end = time.time()
            total_time += (end - start)
            iters += 1
            batch_metrics.append(metrics)

        training_batch_metrics = jax.device_get(batch_metrics)
        training_epoch_metrics = {
            k: np.mean([metrics[k] for metrics in training_batch_metrics])
            for k in training_batch_metrics[0]}
        print(f"🤖 Epoch: {epoch}, "
            f"Train loss: {training_epoch_metrics['loss']:.3f}, "
            f"Train accuracy: {100*training_epoch_metrics['accuracy']:.2f}")

        return state, training_epoch_metrics, [total_time, iters]
    
    
    #@jax.jit
    def eval_step(self, params, batch):
        logits = model.apply({'params': params}, batch['image'])
        return self.compute_metrics(logits, batch['label'])

    def eval_model(self, model, test_ds):
        metrics = self.eval_step(model, test_ds)
        metrics = jax.device_get(metrics)
        eval_summary = jax.tree_map(lambda x: x.item(), metrics)
        print(f"🤖 Test loss: {eval_summary['loss']:.2f},"
            f" Test accuracy: {100*eval_summary['accuracy']:.2f}")
        return eval_summary
    
    def train_model_on_fashionmnist(self):
        RNG = jax.random.PRNGKey(self.config["seed"])
        RNG, init_RNG = jax.random.split(RNG)
        train_ds, test_ds = self.train_ds, self.test_ds

        input_shape = [1, 28, 28, 1] # Batch_size x Width x Height x Channels
        params = model.init(init_RNG, jnp.ones(input_shape))['params']

        # This creates transformation of the parameters w.r.t. Gradient Descent method
        # Note, that stochasticity come from the batches later
        optimizer = optax.sgd(learning_rate=self.config["LEARNING_RATE"])
        if self.method == 'adam':
            optimizer = optax.adam(learning_rate=self.config["LEARNING_RATE"])
        if self.method == 'adagrad':
            optimizer = optax.adagrad(learning_rate=self.config["LEARNING_RATE"])
        if self.method == 'novograd':
            optimizer = optax.novograd(learning_rate=self.config["LEARNING_RATE"])

        state = train_state.TrainState.create(apply_fn=model.apply, 
                                            params=params, 
                                            tx=optimizer)

        # Evaluate on the test set before training epoch
        test_metrics = self.eval_model(state.params, self.test_ds)
        metrics_to_return = dict()
        metrics_to_return['train_accuracy_history'] = list()
        metrics_to_return['test_accuracy_history'] = list()
        metrics_to_return['test_loss_history'] = list()
        metrics_to_return['train_loss_history'] = list()

        for epoch in tqdm(range(1, self.config["N_EPOCHS"])):
            # Use a separate PRNG key to permute image data during shuffling
            RNG, input_RNG = jax.random.split(RNG)
            # Run an optimization step over a training batch
            state, train_metrics, stats = self.train_epoch(state, train_ds, self.config["BATCH_SIZE"], 
                                            epoch, input_RNG)
            # # Evaluate on the test set after each training epoch
            #with open('test.txt', 'w') as file:
            #  print(train_metrics, file=file)
            #self.model = state.params
            test_metrics = self.eval_model(state.params, test_ds)
            
            metrics_to_return['test_loss_history'].append(float(test_metrics['loss']))
            metrics_to_return['train_loss_history'].append(float(train_metrics['loss']))
            metrics_to_return['train_accuracy_history'].append(float(train_metrics['accuracy']))
            metrics_to_return['test_accuracy_history'].append(float(test_metrics['accuracy']))
        #print(metrics_to_return)
        return metrics_to_return
    
    def __del__(self):
        model = CNN()
    


class NNBenchmark:
    # methods: {method: config}. LABEL key in config
    methods: Dict = None
    aval_methods = ['sgd', 'adam', 'novograd', 'adagrad']
    save: bool
    path: str
    def __init__(self, methods: Dict) -> None:
        for label, data in methods.items():
            if data['method'] not in self.aval_methods:
                logging.critical(f'Bad method {data["method"]}. Available methods: {self.aval_methods}')
                exit(1)
        self.methods = methods
    
    def run(self, save: bool = False, path: str = ''):
        if len(path) > 0:
            self.path = path
        result_metrics = dict()

        for label, config in self.methods.items():
            tmp_model = NeuralNetwokTraining(config=config, method=config['method'])
            tmp_metrics = tmp_model.train_model_on_fashionmnist()
            method = config['method']
            config.pop('method')
            result_metrics[label] = tmp_metrics

        if save and len(path) > 0:
            with open(path, 'w') as fp:
                json.dump(result_metrics, fp, indent=4)
        else:
            print(result_metrics)

    def plot(self,
             metrics_to_plot = ['train_acc', 'test_acc', 'train_loss', 'test_loss'],
             data_path='',
             dir_path = '.',
             save: bool = True,
             show: bool = False,
             log: bool = True):
        if len(data_path) == 0:
            data_path = self.path
        
        plotter_ = Plotter(metrics=metrics_to_plot, 
                           data_path=data_path,
                           dir_path=dir_path)
        plotter_._plot_nn_data(save=save, show=show, log=log)


def run_experiment():
    bencmark = NNBenchmark({
        'adam' : {
            "N_EPOCHS": 40,
            "BATCH_SIZE": 25000,
            "Dataset": "FashionMNIST",
            "seed": 282,
            "LEARNING_RATE": 1e-2,
            'method': 'adam'
        },
        'sgd_1' : {
            "N_EPOCHS": 40,
            "BATCH_SIZE": 25000,
            "Dataset": "FashionMNIST",
            "seed": 282,
            "LEARNING_RATE": 1e-2,
            'method': 'sgd'
        },
        'sgd_2' : {
            "N_EPOCHS": 40,
            "BATCH_SIZE": 25000,
            "Dataset": "FashionMNIST",
            "seed": 282,
            "LEARNING_RATE": 0.05,
            'method': 'sgd'
        }
    })

    path_to_save = 'sgd_model_data.json'
    bencmark.run(save=True, path=path_to_save)


def draw():
    from plotter import Plotter

    plot = Plotter(['train_acc', 'test_acc', 'train_loss', 'test_loss'], 
                   data_path='sgd_model_data.json',
                   dir_path='nn_plots')
    plot._plot_nn_data()

def test_local():
    run_experiment()
    draw()


if __name__ == '__main__':
    test_local()


from typing import Callable
import jax
import jax.numpy as jnp
import logging

import sys
import os

from benchmarx.problem import Problem

#sys.path.insert(1, os.path.abspath(os.path.join(os.path.dirname("benchmarx"), "..")))
from benchmarx.defaults import *

class LogLoss(Problem):
    """
    Convex, but not strongly convex.
    y: jnp.array with shape (n,)
    X: jnp.array with shape (n, d), 
        x_i is the ith row of X matrix, 
        x_i shape (d,)
    
    f(w) = 1/n sum i = 1 to n log[1 + exp(-y_i <x_i, w>)]
    """
    X: jnp.array = None
    y: jnp.array = None
    seed = default_seed
    def __init__(self, y = None, X = None, n: int = 2, d: int = 2, info: str='log_loss') -> None:
        self.n = n
        self.d = d
        if y is None:
            self.y = self._generate_y()
        else:
            if y.shape[0] == self.n:
                self.y = y
            else:
                err_msg = f'y must be of shape ({self.n},), not {y.shape}'
                logging.critical(err_msg)
                exit(1)
        if X is None:
            self.X = self._generate_X()
        else:
            if X.shape[0] == self.n and X.shape[1] == self.d:
                self.X = X
            else:
                err_msg = f'X must be of shape ({self.n}, {self.d}), not {X.shape}'
                logging.critical(err_msg)
                exit(1)

        super().__init__(info=info, func=self.f)
        
    
    def f(self, w, *args, **kwargs):
        """
        w: jnp.array with shape (d,)
        """
        if w.shape != (self.d,):
            err_msg = f'Argument must be of shape ({self.d},), not {w.shape}'
            logging.critical(err_msg)
            exit(1)
        
        return sum([jnp.log(1 + jnp.exp(-self.y[i] * self.X[i].T @ w)) for i in range(self.n)]) / self.n


    def _generate_X(self):
        key = jax.random.PRNGKey(self.seed)
        return jax.random.uniform(key, (self.n, self.d))

    def _generate_y(self):
        key = jax.random.PRNGKey(self.seed)
        return jax.random.uniform(key, (self.n))



import pandas as pd
import plotly
import plotly.express as px
import plotly.io as pio

pio.templates.default = "plotly_white"
import plotly.graph_objects as go
import os
import pathlib
import json
import logging
import jax.numpy as jnp
from typing import List, Dict

from benchmarx.defaults import default_plotly_config, default_log_threshold
from benchmarx.metrics import Metrics, CustomMetric

from benchmarx.benchmark_result import BenchmarkResult


class Plotter:
    """
    Class for plotting benchmark results using Plotly.

    Attributes:
        benchmark_result (BenchmarkResult): The benchmark result object.
    """
    benchmark_result: BenchmarkResult

    def __init__(self, benchmark_result) -> None:
        self.benchmark_result = benchmark_result

    def plotly_figure(
        self, dataframe: pd.DataFrame, dropdown_options: List[Dict[str, str]]
    ) -> go.Figure:
        """
        Create a Plotly figure for plotting benchmark results.

        Args:
            dataframe (pd.DataFrame): DataFrame containing benchmark data.
            dropdown_options (List[Dict[str, str]]): Dropdown options for selecting metrics.

        Returns:
            go.Figure: The Plotly figure.
        """
        markers = [
            "circle",
            "square",
            "diamond",
            "cross",
            "x",
            "triangle-up",
            "triangle-down",
            "triangle-left",
            "triangle-right",
            "triangle-ne",
            "triangle-se",
            "triangle-sw",
            "triangle-nw",
        ]
        colors_rgba = [
            "rgba(31, 119, 180,  1)",
            "rgba(255, 127, 14,  1)",
            "rgba(44, 160, 44,   1)",
            "rgba(214, 39, 40,   1)",
            "rgba(148, 103, 189, 1)",
            "rgba(140, 86, 75,   1)",
            "rgba(227, 119, 194, 1)",
            "rgba(127, 127, 127, 1)",
            "rgba(188, 189, 34,  1)",
            "rgba(23, 190, 207,  1)",
        ]
        colors_rgba_faint = [
            "rgba(31, 119, 180,  0.3)",
            "rgba(255, 127, 14,  0.3)",
            "rgba(44, 160, 44,   0.3)",
            "rgba(214, 39, 40,   0.3)",
            "rgba(148, 103, 189, 0.3)",
            "rgba(140, 86, 75,   0.3)",
            "rgba(227, 119, 194, 0.3)",
            "rgba(127, 127, 127, 0.3)",
            "rgba(188, 189, 34,  0.3)",
            "rgba(23, 190, 207,  0.3)",
        ]
        fig = go.Figure()

        # Add traces for each method and each dropdown option
        for i_method, method in enumerate(dataframe["Method"].unique()):
            method_df = dataframe[dataframe["Method"] == method]
            marker = dict(
                symbol=markers[i_method % len(markers)],
                color=colors_rgba[i_method % len(colors_rgba)],
            )
            fillcolor = colors_rgba_faint[i_method % len(colors_rgba_faint)]
            for option in dropdown_options:
                trace_mean = go.Scatter(
                    x=method_df["Iteration"],
                    y=method_df[option["value"] + "_mean"],
                    mode="lines+markers",
                    marker=marker,
                    hovertext=f"{method} - {option['label']}",
                    name=f"{method}",
                    visible=option["value"] == dropdown_options[0]["value"],
                )
                fig.add_trace(trace_mean)
                if not all([val == 0 for val in method_df[option["value"] + "_std"]]):
                    trace_plus_std = go.Scatter(
                        name="mean + std",
                        x=method_df["Iteration"],
                        y=method_df[option["value"] + "_mean"]
                        + method_df[option["value"] + "_std"],
                        mode="lines",
                        line=dict(width=0),
                        showlegend=False,
                        hovertext=f"{method} - {option['label']}_upper",
                        visible=option["value"] == dropdown_options[0]["value"],
                    )
                    fig.add_trace(trace_plus_std)

                    trace_minus_std = go.Scatter(
                        name="mean - std",
                        x=method_df["Iteration"],
                        y=[max(val, default_log_threshold) for val in method_df[option["value"] + "_mean"]- method_df[option["value"] + "_std"]],
                        line=dict(width=0),
                        mode="lines",
                        fillcolor=fillcolor,
                        fill="tonexty",
                        showlegend=False,
                        hovertext=f"{method} - {option['label']}_lower",
                        visible=option["value"] == dropdown_options[0]["value"],
                    )
                    fig.add_trace(trace_minus_std)
        # Update layout
        fig.update_layout(
            updatemenus=[
                {
                    "buttons": [
                        {
                            "method": "update",
                            "label": option["label"],
                            "args": [
                                {
                                    "visible": [
                                        trace.hovertext.endswith(
                                            f" - {option['value']}"
                                        )
                                        or trace.hovertext.endswith(
                                            f" - {option['value']}_upper"
                                        )
                                        or trace.hovertext.endswith(
                                            f" - {option['value']}_lower"
                                        )
                                        for trace in fig.data
                                    ]
                                }
                            ],
                            "args2": [
                                {"yaxis": {"title": option["label"], "type": "log"}}
                            ],
                        }
                        for option in dropdown_options
                    ],
                    "direction": "down",
                    "showactive": True,
                    "x": -0.14,
                    "xanchor": "left",
                    "y": 1.2,
                    "yanchor": "top",
                }
            ],
            xaxis={"title": "Iteration"},
            yaxis={"title": "", "type": "log"},
            title=str(dataframe.T[0]["Problem"]),  # Set your problem title here
        )

        fig.update_layout(
            dragmode="pan",
            title={
                "x": 0.5,
                "xanchor": "center",
            },
        )

        return fig

    def plot(
        self,
        metrics: List[str | CustomMetric] = [],
        plotly_config=default_plotly_config,
        write_html: bool = False,
        path_to_write: str = "",
        include_plotlyjs: str = "cdn",
        full_html: bool = False,
    ) -> None:
        """
        Plot benchmark results using Plotly.

        Args:
            metrics (List[str | CustomMetric]): Metrics to plot.
            plotly_config: Plotly config.
            write_html (bool): If True, write an HTML file.
            path_to_write (str): Path to write the HTML file.
            include_plotlyjs (str): Include Plotly JS in the HTML.
            full_html (bool): Create a full HTML file.

        Returns:
            None
        """

        # Metrics.model_metrics_to_plot are tracking in Benchmark, if the problem
        # inherits from ModelProblem. Thus if the problem inherits from ModelProblem,
        # metrics Metrics.model_metrics_to_plot are always will be "successful", 
        # i.e. this metrics will be in good_str_metrics returned from benchmark_result.get_dataframes
        # (if the problem inherits from ModelProblem), and vice versa: 
        # if the problem does NOT inherit from ModelProblem, Metrics.model_metrics_to_plot
        # will not be contained in good_str_metrics.
        # Therefore, by adding Metrics.model_metrics_to_plot to df_metrics for passing to
        # benchmark_result.get_dataframes, Metrics.model_metrics_to_plot will be plotted in case 
        # if problem inherits from ModelProblem, and will NOT be plotted in case if problem does NOT
        # inherit from ModelProblem.
        dfs, good_str_metrics = self.benchmark_result.get_dataframes(df_metrics=metrics + Metrics.model_metrics_to_plot)
        for _, df in dfs.items():
            dropdown_options = [
                {"label": metric, "value": metric} for metric in good_str_metrics
            ]
            figure = self.plotly_figure(dataframe=df, dropdown_options=dropdown_options)
            figure.show(config=plotly_config)

            if write_html:
                figure.write_html(
                    path_to_write,
                    config=plotly_config,
                    include_plotlyjs="cdn",
                    full_html=False,
                )



from typing import Callable
import jax
import jax.numpy as jnp
import logging

from benchmarx.problem import Problem
from benchmarx.defaults import default_seed

class Rosenbrock(Problem):
    """
    A class describing the Rosenbrock function.

    The Rosenbrock function is a non-convex optimization problem that is often
    used as a benchmark to test optimization algorithms. It is characterized
    by a narrow, parabolic shaped valley that contains the global minimum.

    The function is defined as:
    f(x) = sum(100 * (x[i+1] - x[i]^2)^2 + (1 - x[i])^2) for i in range(n-1)

    where n is the dimensionality of the problem and x is the input vector.
    """
    n: int = 2      # problem dimensionality
    seed = default_seed
    def __init__(self, n: int = 2, info: str = 'Rosenbrock') -> None:
        """
        Initialize the Rosenbrock instance.

        Args:
            n (int, optional): Problem dimensionality.
            info (str, optional): Brief information about the problem.
        """
        self.info = info
        self.n = n
        self.x_opt = jnp.ones(self.n)
        self.f_opt = 0.0
        super().__init__(info=info, func=self.f, x_opt=self.x_opt)
    

    def f(self, x, *args, **kwargs):
        """
        Calculate the Rosenbrock function value.

        Args:
            x (Any): Input value for the function.

        Returns:
            Any: The Rosenbrock function value.
        """
        if x.shape[0] != self.n:
            err_msg = f'Wrong x shape: {x.shape}. Expected: ({self.n},)'
            logging.critical(err_msg)
            exit(1)
        return sum([100 * (x[i+1] - x[i]**2)**2 + (1 - x[i])**2 for i in range(self.n - 1)])



from setuptools import find_packages, setup


with open("README.md", "r") as f:
    long_description = f.read()


setup(
    name="benchmarx",
    version="0.0.11",
    description="Tools for benchmarking optimization methods",
    #package_dir={"" : "benchmarx"},
    #py_modules=['benchmarx'],
    packages=find_packages(),
    #packages=find_packages(where="benchmarx"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AlexToW/benchmarx",
    author="AlexToW",
    author_email="31salex31@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "hatchling",
        "jax",
        "jaxopt >= 0.6",
        "typing-extensions >= 4.5.0",
        "wandb >= 0.14.2",
        "pandas == 1.5.3",
        "plotly >= 5.14.1",
        "flax >= 0.6.10",
        "tensorflow_datasets"
    ],
    python_requires=">=3.7",
)



import json
import logging
import sys
import uuid

import jax
import jax.numpy as jnp
import pandas as pd

from benchmarx.metrics import Metrics, CustomMetric
from benchmarx.problem import Problem
#from benchmarx.plotter import Plotter 
from benchmarx.quadratic_problem import QuadraticProblem
from benchmarx.defaults import default_plotly_config

from typing import List, Dict, Tuple


class BenchmarkResult:
    """
    BenchmarkResult class contains benchmarking results.
    """
    methods: List[str] = None
    problem: Problem = None  # the Problem on which the benchmarking was performed
    data: Dict[Problem, Dict[str, Dict[str, List[any]]]] = None

    def __init__(
        self,
        problem: Problem = None,
        methods: List[str] = None,
        data: Dict[Problem, Dict[str, Dict[str, List[any]]]] = None,
    ) -> None:
        """
        Initialize BenchmarkResult instance.

        :param problem: The Problem on which the benchmarking was performed.
        :param methods: List of method names.
        :param data: Dictionary containing benchmarking data.
        """
        self.problem = problem
        self.methods = methods
        self.data = data
    
    @staticmethod
    def _convert(val):
        """
        Converts val from str to appropriate type.

        :param val: Value to convert.
        :return: Converted value.
        """

        if isinstance(val, float) or isinstance(val, jnp.ndarray):
            return val

        if isinstance(val, List):
            if len(val) > 0 and isinstance(val[0], str):
                if val[0][0] == "[" and val[0][-1] == "]":
                    # val is like ['[2. 1.]', '[7.5 8.]']
                    res = list()
                    for item in val:
                        item = item.replace("\n", "")
                        tmp = jnp.array(
                            [float(x) for x in item[1:-1].split(" ") if len(x) > 0]
                        )
                        res.append(tmp)
                    return res
                else:
                    # val is like ['4.5', '-0.1']
                    return [float(x) for x in val]

        elif isinstance(val, str):
            val = val.replace("\n", "")
            if val[0] == "[" and val[-1] == "]":
                # val is like '[2. 1.]'
                return jnp.array(
                    [float(x) for x in (val[1:-1]).split(" ") if len(x) > 0]
                )
            else:
                flag = True
                try:
                    tmp = int(val)
                except:
                    flag = False
                if flag:
                    return int(val)

                flag = True
                try:
                    tmp = float(val)
                except:
                    flag = False

                if flag:
                    return float(val)

                return val
        else:
            # something went wrong
            logging.critical(f"Can't convert {val} of type {type(val)}")
            return "wtf"

    @staticmethod
    def _matrix_from_str(A_str: str):
        """
        Converts string representation of a matrix to a numpy array.

        :param A_str: String representation of the matrix.
        :return: Numpy array representing the matrix.

        A_str in format:
        "[[0.96531415 0.84779143 0.72762513]\n [0.31114805 0.03425407 0.31510842]\n [0.12594318 0.42591357 0.8050107 ]]"
        """
        pre_raws = A_str.split("]")
        pre_raws[0] = pre_raws[0][1:]
        pre_raws = [val.strip() + "]" for val in pre_raws if len(val) > 0]
        raws = [raw[1:-1].strip().replace("\n", "") for raw in pre_raws]
        A = jnp.array([jnp.fromstring(raw, sep=" ") for raw in raws])
        return A

    @classmethod
    def load(cls, path: str) -> None:
        """
        Loads benchmark data from a JSON file.

        :param path: Path to the JSON file.
        :return: Loaded BenchmarkResult instance.
        """
        raw_data = dict()
        with open(path) as json_file:
            raw_data = json.load(json_file)

        good_data = dict()
        problem_obj = None
        methods = list()
        for problem, problem_dict in raw_data.items():
            x_opt = None
            A = None
            b = None
            f_opt = None
            if "A" in raw_data[problem]:
                A = BenchmarkResult._matrix_from_str(raw_data[problem]["A"])
                raw_data[problem].pop("A")
            if "b" in raw_data[problem]:
                b = jnp.fromstring(raw_data[problem]["b"][1:-1], sep=" ")
                raw_data[problem].pop("b")
            if "x_opt" in raw_data[problem]:
                x_opt = BenchmarkResult._convert(raw_data[problem]["x_opt"])
                raw_data[problem].pop("x_opt")
            if "f_opt" in raw_data[problem]:
                f_opt = BenchmarkResult._convert(raw_data[problem]["f_opt"])
                raw_data[problem].pop("f_opt")

            problem_dict_good = dict()
            for method, method_dict in problem_dict.items():
                methods.append(str(method))
                hyperparams_good = dict()
                runs_good = dict()
                for field, field_dict in method_dict.items():
                    # field is 'hyperparams' or 'runs'
                    if field == "hyperparams":
                        for hyperparam, val in field_dict.items():
                            hyperparams_good[str(hyperparam)] = BenchmarkResult._convert(val)
                    if field == "runs":
                        for run_num, run_dict in field_dict.items():
                            tmp_run_good = dict()
                            for metric, metric_val in run_dict.items():
                                tmp_run_good[str(metric)] = BenchmarkResult._convert(metric_val)
                            runs_good[str(run_num)] = tmp_run_good
                method_dict_good = {"hyperparams": hyperparams_good, "runs": runs_good}
                problem_dict_good[str(method)] = method_dict_good

            good_data[str(problem)] = problem_dict_good
            if x_opt is not None:
                good_data[str(problem)]["x_opt"] = x_opt
            if A is not None:
                good_data[problem]["A"] = A
            if b is not None:
                good_data[problem]["b"] = b
            if f_opt is not None:
                good_data[problem]["f_opt"] = f_opt
            
            if A is not None and b is not None:
                problem_obj = QuadraticProblem(
                    n=A.shape[0],
                    A=A,
                    b=b,
                    info=problem
                )
            if x_opt is not None:
                problem_obj.x_opt = x_opt
            if f_opt is not None:
                problem_obj.f_opt = f_opt

        return cls(
            problem=problem_obj,
            methods=methods,
            data=good_data
        )

    def save(self, path: str) -> None:
        """
        Saves benchmarking data to a JSON file.

        :param path: Path to save the JSON file.
        """
        # data: dict[Problem, dict[str, dict[str, list[any]]]]
        data_str = dict()
        for problem, ddict in self.data.items():
            tmp2 = dict()
            for method, dddict in ddict.items():
                tmp1 = dict()
                hyper_dict = dict()
                runs_dict = dict()
                for field, d in dddict.items():
                    # field is 'hyperparams' or 'runs'.
                    if str(field) == 'hyperparams':
                        hyper_dict = {key: str(val) for key, val in d.items()}

                    if str(field) == 'runs':
                        for run_num, run_dict in d.items():
                            run_num_dict_str = dict()
                            metric_lst_str = list()
                            for metric, metric_lst in run_dict.items():
                                metric_lst_str = [str(val) for val in metric_lst]
                                if len(metric_lst_str) > 1:
                                    run_num_dict_str[str(metric)] = metric_lst_str
                                else:
                                    run_num_dict_str[str(metric)] = metric_lst_str[0]
                            runs_dict[str(run_num)] = run_num_dict_str
                    
                    tmp1 = {'hyperparams': hyper_dict, 'runs': runs_dict}
                tmp2[method] = tmp1
            data_str[str(problem)] = tmp2

            if self.problem is not None:
                if self.problem.x_opt is not None:
                    data_str[str(problem)]['x_opt'] = str(self.problem.x_opt)

                if self.problem.f_opt is not None:
                    data_str[str(problem)]['f_opt'] = str(self.problem.f_opt)
                
                if isinstance(problem, QuadraticProblem):
                    jnp.set_printoptions(threshold=sys.maxsize)
                    data_str[str(problem)]['A'] = jnp.array_str(self.problem.A)
                    data_str[str(problem)]['b'] = str(self.problem.b)

        with open(path, "w") as file:
            json.dump(data_str, file, indent=2)

    def get_dataframes(self, df_metrics: List[str | CustomMetric]) -> Tuple[Dict[str, pd.DataFrame], List[str]]:
        """
        Create DataFrames from rows.

        :param df_metrics: List of metrics to put in columns.
        :return: Tuple of DataFrames and list of successfully calculated metrics.

        Create DataFrame form rows.
        Returns tuple of:
            [0] dictionary {problem: problem's DataFrame}.
            [1] A list of metrics that have been successfully calculated.
        Only successfully calculated metrics will be plotted.
        df_metrics -- metrics to put in columns, subset of
        Metrics.metrics_to_plot or your CustomMetric objects
        """
        result_dict = {}
        problem_rows = {}

        success_metrics = []

        for problem, problem_data in self.data.items():
            x_opt = None
            f_opt = None
            if "A" in problem_data:
                problem_data.pop("A")
            if "b" in problem_data:
                problem_data.pop("b")

            # after these manipulations either x_opt is not None, or it can no longer be counted
            if "x_opt" in problem_data:
                x_opt = self._convert(problem_data["x_opt"])
                problem_data.pop("x_opt")
            elif isinstance(self.problem, Problem) and hasattr(self.problem, "x_opt"):
                x_opt = self.problem.x_opt

            # after these manipulations either f_opt is not None, or it can no longer be counted
            if "f_opt" in problem_data:
                f_opt = self._convert(problem_data["f_opt"])
                problem_data.pop("f_opt")
            elif isinstance(self.problem, Problem):
                if hasattr(self.problem, "f_opt"):
                    f_opt = self.problem.f_opt
                elif hasattr(self.problem, "x_opt"):
                    f_opt = self.problem.f(self.problem.x_opt)

            rows = []

            for method, method_data in problem_data.items():
                if "hyperparams" in method_data and "runs" in method_data:
                    hyperparams = method_data["hyperparams"]
                    runs = method_data["runs"]
                    nit = int(runs["run_0"]["nit"][0])

                    for iteration in range(nit):
                        row = {
                            "Problem": problem,
                            "Method": method_data["hyperparams"]["label"],
                            "Hyperparameters": hyperparams,
                            "Iteration": iteration,
                        }
                        if x_opt is not None:
                            row["x_opt"] = x_opt
                        if f_opt is not None:
                            row["f_opt"] = f_opt

                        to_means_stds = {}
                        for run, run_data in runs.items():
                            run_num = int(run[4:])
                            for df_metric in df_metrics:
                                df_metric_key = str(df_metric)  # string representation of metric (label)
                                if df_metric_key not in success_metrics:
                                    success_metrics.append(df_metric_key)
                                custom_df_metric_flag = isinstance(df_metric, CustomMetric)
                                if isinstance(df_metric, str):
                                    if df_metric not in Metrics.metrics_to_plot and df_metric not in Metrics.model_metrics_to_plot:
                                        logging.warning(
                                            f"Unknown metric '{df_metric}'. Use CustomMetric to specify your own metric."
                                        )
                                        if df_metric_key in success_metrics:
                                            success_metrics.remove(df_metric_key)

                                val = None
                                if df_metric_key in run_data.keys():
                                    val = run_data[df_metric_key][iteration]
                                elif custom_df_metric_flag:
                                    try:
                                        val = df_metric.func(
                                            run_data["x"][iteration]
                                        )
                                    except:
                                        logging.warning(
                                            msg=f"Something went wrong while calculating the CustomMetric '{df_metric_key}'."
                                        )
                                        if df_metric_key in success_metrics:
                                            success_metrics.remove(df_metric_key)
                                elif df_metric_key == "x_gap":
                                    if x_opt is not None:
                                        val = float(jnp.linalg.norm(run_data["x"][iteration] - x_opt))
                                    else:
                                        logging.warning(
                                            msg=f"Cannot calculate x_gap because there is no information about x_opt."
                                        )
                                        if df_metric_key in success_metrics:
                                            success_metrics.remove(df_metric_key)
                                elif df_metric_key == "f_gap":
                                    if f_opt is not None:
                                        if "f" in run_data.keys():
                                            val = jnp.abs(run_data["f"][iteration] - f_opt)
                                        elif isinstance(self.problem, Problem):
                                            val = jnp.abs(self.problem.f(run_data["x"][iteration]) - f_opt)
                                        else:
                                            logging.warning(
                                                msg="Cannot calculate f_gap because there is impossible to compute f_opt."
                                            )
                                            if df_metric_key in success_metrics:
                                                success_metrics.remove(df_metric_key)
                                    else:
                                        logging.warning(
                                            msg="Cannot calculate f_gap because there is impossible to compute f_opt."
                                        )
                                        if df_metric_key in success_metrics:
                                            success_metrics.remove(df_metric_key)
                                elif df_metric_key == "grad_norm":
                                    if "grad" in run_data.keys():
                                        val = float(jnp.linalg.norm(run_data["grad"][iteration]))
                                    elif isinstance(self.problem, Problem):
                                        if hasattr(self.problem, "grad"):
                                            val = float(jnp.linalg.norm(self.problem.grad(run_data["x"][iteration])))
                                        else:
                                            val = float(jnp.linalg.norm(jax.grad(self.problem.f)(run_data["x"][iteration])))
                                    else:
                                        logging.warning(
                                            msg="Cannot calculate 'grad_norm' because there is no information about grad."
                                        )
                                        if df_metric_key in success_metrics:
                                            success_metrics.remove(df_metric_key)
                                elif df_metric_key == "x_norm":
                                    val = float(jnp.linalg.norm(run_data["x"][iteration]))
                                elif df_metric_key == "f":
                                    if isinstance(self.problem, Problem):
                                        val = self.problem.f(run_data["x"][iteration])
                                    else:
                                        logging.warning(
                                            msg="Cannot calculate 'f' metric because there is no information about problem."
                                        )
                                        if df_metric_key in success_metrics:
                                            success_metrics.remove(df_metric_key)
                                elif df_metric_key == "relative_x_gap":
                                    if x_opt is not None:
                                        if jnp.linalg.norm(x_opt) != 0:
                                            val = jnp.linalg.norm(run_data["x"][iteration] - x_opt) / jnp.linalg.norm(x_opt)
                                        else:
                                            logging.warning(
                                                msg="Cannot calculate 'relative_x_gap' metric because ||x_opt|| = 0."
                                            )
                                            if df_metric_key in success_metrics:
                                                success_metrics.remove(df_metric_key)
                                    else:
                                        logging.warning(
                                            msg="Cannot calculate 'relative_x_gap' metric because there is no information about x_opt."
                                        )
                                        if df_metric_key in success_metrics:
                                            success_metrics.remove(df_metric_key)
                                
                                elif df_metric_key == "relative_f_gap":
                                    if f_opt is not None:
                                        if f_opt != 0:
                                            if "f" in run_data.keys():
                                                val = jnp.abs(run_data["f"][iteration] - f_opt) / jnp.abs(f_opt)
                                            elif isinstance(self.problem, Problem):
                                                f_val = self.problem.f(run_data["x"][iteration])
                                                val = jnp.abs(f_val - f_opt) / jnp.abs(f_opt)
                                            else:
                                                logging.warning(
                                                    msg="Cannot calculate 'relative_f_gap' metric because there is no information about f."
                                                )
                                                if df_metric_key in success_metrics:
                                                    success_metrics.remove(df_metric_key)
                                        else:
                                            logging.warning(
                                                msg="Cannot calculate 'relative_f_gap' metric because f_opt = 0."
                                            )
                                            if df_metric_key in success_metrics:
                                                success_metrics.remove(df_metric_key)
                                    else:
                                        logging.warning(
                                            msg="Cannot calculate 'relative_f_gap' metric because there is no information about f_opt."
                                        )
                                        if df_metric_key in success_metrics:
                                            success_metrics.remove(df_metric_key)

                                row[df_metric_key + "_" + str(run_num)] = val
                                if df_metric_key in to_means_stds.keys():
                                    to_means_stds[df_metric_key].append(val)
                                else:
                                    to_means_stds[df_metric_key] = [val]

                        for metric, val1 in to_means_stds.items():
                            row[metric + "_mean"] = jnp.mean(jnp.array(val1), axis=0)
                            row[metric + "_std"] = jnp.std(jnp.array(val1), axis=0)

                        rows.append(row)
            problem_rows[problem] = rows

        for problem, rows in problem_rows.items():
            result_dict[problem] = pd.DataFrame(rows)

        return result_dict, success_metrics

    def plot(
        self,
        metrics: List[str | CustomMetric] = [],
        plotly_config=default_plotly_config,
        write_html: bool = False,
        path_to_write: str = "",
        include_plotlyjs: str = "cdn",
        full_html: bool = False
    ) -> None:
        """
        Plot benchmarking data.

        :param metrics: List of metrics to plot.
        :param plotly_config: Plotly configuration.
        :param write_html: Flag to write HTML plot.
        :param path_to_write: Path to write HTML plot.
        :param include_plotlyjs: Option for including Plotly JS.
        :param full_html: Flag for full HTML plot.
        """
        from benchmarx.plotter import Plotter 
        plotter = Plotter(
            benchmark_result=self
        )
        plotter.plot(
            metrics=metrics,
            plotly_config=plotly_config,
            write_html=write_html,
            path_to_write=path_to_write,
            include_plotlyjs=include_plotlyjs,
            full_html=full_html
        )



import jax
import sys
import os
import scipy.linalg as la
import logging
import jax.numpy as jnp


from benchmarx.problem import Problem

#sys.path.insert(1, os.path.abspath(os.path.join(os.path.dirname("benchmarx"), "..")))
from benchmarx.defaults import *


class QuadraticProblem(Problem):
    """
    A class describing an unconstrained quadratic problem:
    0.5 x^TAx^T + b^Tx, where the matrix A is positive defined, x \in R^n
    """

    n: int = 1  # problem dimensionality
    A = None  # A-matrix: np.array of shape (n,n)
    b = None  # b-vector: np.array of shape (n,)
    seed = default_seed
    def __init__(self, n: int = 2, A=None, b=None, mineig=0, maxeig=1, info: str = "Quadratic problem"):
        """
        Initialize the QuadraticProblem instance.

        Args:
            n (int): Problem dimensionality.
            A (np.ndarray, optional): A-matrix of shape (n,n).
            b (np.ndarray, optional): b-vector of shape (n,).
            mineig (float, optional): Minimum eigenvalue for random matrix A.
            maxeig (float, optional): Maximum eigenvalue for random matrix A.
            info (str, optional): Brief information about the problem.
        """
        self.n = n

        if A is None:
            self.A = self.__get_random_matrix(self.n, maxeig=maxeig, mineig=mineig)
        else:
            self.A = A

        if b is None:
            self.b = self.__get_random_vector(self.n)
        else:
            self.b = b

        # func = lambda x: 0.5 * x.T @ self.A @ x + self.b.T @ x
        super().__init__(info=info, func=self.f)
        if jnp.linalg.det(self.A + self.A.T) != 0:
            self.x_opt = -2 * jnp.linalg.inv(self.A + self.A.T) @ self.b
            self.f_opt = self.f(self.x_opt)

    def f(self, x, *args, **kwargs):
        """
        Calculate the quadratic function value.

        Args:
            x (Any): Input value for the function.

        Returns:
            Any: The quadratic function value.
        """
        x = jnp.array(x)
        return 0.5 * x.T @ self.A @ x + self.b.T @ x

    def __get_random_matrix(self, n: int = 2, mineig=0, maxeig=1):
        """
        Returns a positive defined matrix of size (n, n)
        with eigenvalues in [mineig, maxeig]
        """
        key = jax.random.PRNGKey(self.seed)

        lambdas = list()  # eigenvalues of matrix to generate

        if n == 1:
            if mineig == maxeig:
                lambdas = [mineig]
            else:
                logging.critical(msg="It is impossible to create a matrix of shape (1,1) with two different eigenvalues.")
        if n == 2:
            lambdas = [mineig, maxeig]
        else:
            lambdas = jax.random.uniform(key, minval=mineig, maxval=maxeig, shape=(n-2,))
            lambdas = lambdas.tolist() + [mineig, maxeig]

        A = jnp.diag(jnp.array(lambdas))
        q, _ = la.qr(jax.random.uniform(key, (n, n)))
        A = q.T @ A @ q
        return A

    def __get_random_vector(self, n: int = 2):
        """
        Returns a random vector: np.ndarray of shape (n,)
        """
        key = jax.random.PRNGKey(self.seed)
        return jax.random.uniform(key, (n,))


def local_test():
    n = 3
    qp = QuadraticProblem(n=n)
    print(qp.info, "\n", qp.A)


# local_test()



from typing import Callable
import jax
import sys
import os
import jax.numpy as jnp
import pandas as pd

from benchmarx.problem import Problem
from benchmarx.defaults import default_seed


class QuadraticProblemRealData(Problem):
    """
    World bank data https://data.worldbank.org/
    """
    A = None
    b = None
    n: int = 1
    seed = default_seed
    def __init__(self, info: str = 'QP_World_Bank_data') -> None:
        dataset = pd.read_csv('https://raw.githubusercontent.com/MerkulovDaniil/sber21_fmin/sources/data/world_bank_data.csv')
        y_variable = 'Import'
        x_variable = 'Export'

        clean_data = dataset[[x_variable, y_variable]].dropna()
        
        x_data = jnp.array(clean_data[x_variable].values)
        y_data = jnp.array(clean_data[y_variable].values)
        self.n = x_data.shape[0]
        self.A = jnp.array([float(x_data.T @ x_data)]).reshape((1,1))
        self.b = jnp.array([float(-2 * x_data.T @ y_data)]).reshape((1,))

        B = jnp.vstack([x_data, jnp.ones(len(x_data))]).T
        x_opt, _ = jnp.linalg.lstsq(B, y_data, rcond=None)[0]
        self.x_opt = jnp.array([float(x_opt)])
        self.f_opt = self.f(self.x_opt)
        
        super().__init__(info=info, func=self.f, x_opt=self.x_opt)

    def f(self, w, *args, **kwargs):
        """
        w: jnp.array
        """
        return w.T @ self.A @ w + self.b.T @ w

